# Copyright (c) The University of Edinburgh 2014
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
The dispel4py workflow graph.
'''

import networkx as nx
import inspect
import sys

from dispel4py.core import GenericPE


class WorkflowNode:
    '''
    Wrapper class for workflow nodes - wraps around general subclasses
    of classes denoting PEs, that is GenericPEs.
    '''

    # Supported types of workflow nodes:
    WORKFLOW_NODE_PE = 0
    WORKFLOW_NODE_FN = 1
    WORKFLOW_NODE_CP = 2
    node_counter = 0

    def __init__(self, o):
        self.obj = o
        self.outputs = []
        self.inputs = []

        if inspect.isroutine(o):  # it's a 'function'
            self.nodeType = self.WORKFLOW_NODE_FN
        elif isinstance(o, GenericPE):
            o.id = o.name + str(WorkflowNode.node_counter)
            WorkflowNode.node_counter += 1
            self.nodeType = self.WORKFLOW_NODE_PE
            for i in o.inputconnections.values():
                # empty for the time being - only the index matters
                self.inputs.append({})
            for i in o.outputconnections.values():
                self.outputs.append({})
        elif isinstance(o, WorkflowGraph):
            self.nodeType = self.WORKFLOW_NODE_CP
            try:
                for i in o.inputmappings:
                    self.inputs.append({})
            except AttributeError:
                pass
            try:
                for i in o.outputmappings:
                    self.outputs.append({})
            except AttributeError:
                pass
        else:
            sys.stderr.write('Error: Unknown type of object passed as a \
                              Workflow Node: %s\n' % type(o))
            raise Exception("Unknown type of object passed as a \
                             Workflow Node: %s" % type(o))

    def getContainedObject(self):
        ''' Returns the wrapped PE or function. '''
        return self.obj


# Used as attribute names
FROM_CONNECTION = 'from_connection'
TO_CONNECTION = 'to_connection'
DIRECTION = 'direction'


class WorkflowGraph(object):
    """
    A graph representing the workflow and related methods
    """

    def __init__(self):
        self.graph = nx.Graph()
        self.objToNode = {}

    def add(self, n):
        '''
        Adds node n, which must be an instance of
        :py:class:`dispel4py.core.GenericPE`, and returns the created workflow
        node.

        :rtype: WorkflowNode
        '''
        nd = WorkflowNode(n)
        self.graph.add_node(nd)
        self.objToNode[n] = nd
        return nd

    def connect(self, fromNode, fromConnection, toNode, toConnection):
        '''
        Connect the two given nodes from the given output to the given input.
        If the nodes are not in the graph, they will be added.

        :param fromNode: the source PE of the connection
        :param fromConnection: the name of the output of the source node
        'fromNode'
        :type fromConnection: String
        :param toNode: the destination PE of the connection
        :param toConnection: the name of the input of the destination node
        'toNode'
        :type toConnection: String
        '''

        if fromNode not in self.objToNode:
            self.add(fromNode)
        if toNode not in self.objToNode:
            self.add(toNode)

        fromWfNode = self.objToNode[fromNode]
        toWfNode = self.objToNode[toNode]

        if self.graph.has_edge(fromWfNode, toWfNode):
            self.graph[fromWfNode][toWfNode]['ALL_CONNECTIONS']\
                .append((fromConnection, toConnection))
        else:
            self.graph.add_edge(fromWfNode, toWfNode,
                                **{'FROM_CONNECTION': fromConnection,
                                   'TO_CONNECTION': toConnection,
                                   'DIRECTION': (fromNode, toNode),
                                   'ALL_CONNECTIONS': [
                                       (fromConnection, toConnection)]})

    def getContainedObjects(self):
        nodes = [node.getContainedObject() for node in self.graph.nodes()]
        return sorted(nodes, key=lambda x: x.id)

    def propagate_types(self):
        '''
        Propagates the types throughout the graph by retrieving the output
        types from each node, starting from the root, and providing them to
        connected consumers.
        '''
        visited = set()
        for node in self.graph.nodes():
            if node not in visited:
                self.__assign_types(node, visited)

    def __assign_types(self, node, visited):
        pe = node.getContainedObject()
        inputTypes = {}
        for edge in self.graph[node].values():
            if pe == edge['DIRECTION'][1]:
                # pe is destination so look up the types produced by sources
                source = edge['DIRECTION'][0]
                sourceNode = self.objToNode[source]
                if sourceNode not in visited:
                    self.__assign_types(sourceNode, visited)
                inType = source.getOutputTypes()[edge['FROM_CONNECTION']]
                inputTypes[edge['TO_CONNECTION']] = inType
        pe.setInputTypes(inputTypes)
        visited.add(node)
        # print "%s: Assigned inputs = %s, \
        # received outputs = %s" % \
        # (pe.__class__.__name__, inputTypes, pe.getOutputTypes())

    def flatten(self):
        '''
        Subgraphs contained within composite PEs are added to the top level
        workflow.
        '''
        hasComposites = True
        toRemove = set()
        while hasComposites:
            hasComposites = False
            toRemove = set()
            for node in self.graph.nodes():
                if node.nodeType == WorkflowNode.WORKFLOW_NODE_CP:
                    hasComposites = True
                    toRemove.add(node)
                    wfGraph = node.getContainedObject()
                    subgraph = wfGraph.graph
                    self.graph.add_nodes_from(subgraph.nodes(data=True))
                    self.graph.add_edges_from(subgraph.edges(data=True))
                    self.objToNode.update(wfGraph.objToNode)
                    for inputname in wfGraph.inputmappings:
                        toPE, toConnection = wfGraph.inputmappings[inputname]
                        edge = None
                        fromPE, fromConnection = None, None
                        for e in self.graph[node].values():
                            if wfGraph == e['DIRECTION'][1] \
                                    and inputname == e['TO_CONNECTION']:
                                fromPE = e['DIRECTION'][0]
                                fromConnection = e['FROM_CONNECTION']
                                edge = self.objToNode[fromPE],\
                                    self.objToNode[wfGraph]
                                break
                        if edge is not None:
                            self.connect(fromPE, fromConnection,
                                         toPE, toConnection)
                    for outputname in wfGraph.outputmappings:
                        fromPE, fromConnection = \
                            wfGraph.outputmappings[outputname]
                        destinations = []
                        for e in self.graph[node].values():
                            if wfGraph == e['DIRECTION'][0] \
                                    and outputname == e['FROM_CONNECTION']:
                                toPE = e['DIRECTION'][1]
                                toConnection = e['TO_CONNECTION']
                                destinations.append((toPE, toConnection))
                        for (toPE, toConnection) in destinations:
                            # print 'connecting output %s.%s' %\
                            # (toPE, toConnection)
                            self.connect(fromPE, fromConnection,
                                         toPE, toConnection)
            self.graph.remove_nodes_from(toRemove)


def _create_dot(graph, instanceNames={}, counter=0):
    dot = ''
    # assign unique names
    for node in graph.graph.nodes():
        try:
            name = node.getContainedObject().id, counter
        except:
            name = node.getContainedObject().__class__.__name__, counter
        instanceNames[node] = name
        counter += 1

    # now add all the nodes and their input and output connections
    cluster_index = 0
    for node in graph.graph.nodes():
        pe = node.getContainedObject()
        if isinstance(pe, WorkflowGraph):
            dot += _create_cluster(pe, cluster_index, instanceNames, counter)
            cluster_index += 1
            continue

        name, index = instanceNames[node]
        dot += name + str(index) + "[label=\"{ "
        # add inputs
        inputNames = []
        outputNames = []
        for edge in graph.graph[node].values():
            if pe == edge['DIRECTION'][1]:
                inputName = edge['TO_CONNECTION']
                dotName = "<in_" + inputName + ">" + inputName
                if dotName not in inputNames:
                    inputNames.append(dotName)
            else:
                outputName = edge['FROM_CONNECTION']
                dotName = "<out_" + outputName + ">" + outputName
                if dotName not in outputNames:
                    outputNames.append(dotName)

        if inputNames:
            dot += '{' + ' | '.join(inputNames) + '} | '
        dot += name
        if outputNames:
            dot += ' | {' + ' | '.join(outputNames) + '}'
        dot += " }\"];\n"

    # connect the inputs and outputs
    for node in graph.graph.nodes():
        pe = node.getContainedObject()
        for edge in graph.graph[node].values():
            if pe == edge['DIRECTION'][0]:
                if isinstance(pe, WorkflowGraph):
                    inner_source, source_output = \
                        pe.outputmappings[edge['FROM_CONNECTION']]
                    node = pe.objToNode[inner_source]
                else:
                    source_output = edge['FROM_CONNECTION']
                # pe is the source so look up the connected destination
                dest = edge['DIRECTION'][1]
                if isinstance(dest, WorkflowGraph):
                    inner_dest, dest_input = \
                        dest.inputmappings[edge['TO_CONNECTION']]
                    destNode = dest.objToNode[inner_dest]
                else:
                    destNode = graph.objToNode[dest]
                    dest_input = edge['TO_CONNECTION']
                dot += '%s%s' % instanceNames[node] + ':out_' + source_output
                dot += ' -> '
                dot += '%s%s' % instanceNames[destNode] + ':in_' + dest_input
                dot += ';\n'
    return dot


def _create_cluster(graph, index, instanceNames, counter):
    dot = 'subgraph cluster_%s {\n' % index
    try:
        # names for composite PEs are optional
        dot += 'label = "%s";' % graph.name
    except:
        pass
    dot += 'style=filled;\n'
    dot += 'color=lightgrey;\n'
    if index % 2:
        dot += 'fillcolor=lightgrey;\n'
    dot += _create_dot(graph, instanceNames, counter)
    dot += '}\n'
    return dot


def draw(graph):
    '''
    Creates a representation of the workflow graph in the dot language.
    '''
    dot = 'digraph request\n{\nnode [shape=Mrecord, \
           style=filled, fillcolor=white];\n'
    dot += _create_dot(graph)
    dot += '}\n'
    return dot


def drawDot(graph):
    '''
    Draws the workflow as a graph and creates a PNG image using graphviz dot.
    '''
    from subprocess import Popen, PIPE
    dot = draw(graph)
    p = Popen(['dot', '-T', 'png'], stdout=PIPE, stdin=PIPE, stderr=PIPE)
    stdout, stderr = p.communicate(dot.encode('utf-8'))
    return stdout


def getConnectedInputs(node, graph):
    names = []
    for edge in graph.edges(node, data=True):
        direction = edge[2]['DIRECTION']
        dest = direction[1]
        dest_input = edge[2]['TO_CONNECTION']
        if dest == node.getContainedObject():
            names.append(dest_input)
    return names



# Copyright (c) The University of Edinburgh 2014
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
This is a dispel4py graph which produces two pipeline workflows which are
unconnected.

.. image:: /images/unconnected_pipeline.png

It can be executed with MPI and STORM.

* MPI: Please, locate yourself into the dispel4py directory.

    Execute the MPI mapping as follows::

        mpiexec -n <number mpi_processes> dispel4py mpi <module|module file>\\
            [-a name_dispel4py_graph]\\
            [-f file containing the input dataset in JSON format]\\
            [-i number of iterations/runs']\\
            [-s]

    The argument '-s' forces to run the graph in a simple processing, which
    means that the first node of the graph will be executed in a process, and
    the rest of nodes will be executed in a second process.
    When [-i number of interations/runs] is not indicated, the graph is
    executed once by default.

    For example::

        mpiexec -n 12 dispel4py mpi \\
            dispel4py.examples.graph_testing.unconnected_pipeline

    .. note::

        Each node in the graph is executed as a separate MPI process.
        This graph has 12 nodes. For this reason we need at least 12 MPI
        processes to execute it.

    Output::

        Processing 1 iteration.
        Processes: {'TestProducer0': [9], 'TestProducer6': [6], \
'TestOneInOneOut9': [7], 'TestOneInOneOut8': [8], 'TestOneInOneOut7': [11], \
'TestOneInOneOut5': [1], 'TestOneInOneOut4': [2], 'TestOneInOneOut3': [4], \
'TestOneInOneOut2': [0], 'TestOneInOneOut1': [10], 'TestOneInOneOut11': [3], \
'TestOneInOneOut10': [5]}
        TestProducer6 (rank 6): Processed 1 iteration.
        TestProducer0 (rank 9): Processed 1 iteration.
        TestOneInOneOut1 (rank 10): Processed 1 iteration.
        TestOneInOneOut7 (rank 11): Processed 1 iteration.
        TestOneInOneOut2 (rank 0): Processed 1 iteration.
        TestOneInOneOut8 (rank 8): Processed 1 iteration.
        TestOneInOneOut9 (rank 7): Processed 1 iteration.
        TestOneInOneOut10 (rank 5): Processed 1 iteration.
        TestOneInOneOut3 (rank 4): Processed 1 iteration.
        TestOneInOneOut4 (rank 2): Processed 1 iteration.
        TestOneInOneOut11 (rank 3): Processed 1 iteration.
        TestOneInOneOut5 (rank 1): Processed 1 iteration.
'''

from dispel4py.examples.graph_testing import testing_PEs as t
from dispel4py.workflow_graph import WorkflowGraph


def testPipeline(graph):
    '''
    Creates a pipeline and adds it to the given graph.

    :rtype: the modified graph
    '''
    prod = t.TestProducer()
    prev = prod
    for i in range(5):
        cons = t.TestOneInOneOut()
        graph.connect(prev, 'output', cons, 'input')
        prev = cons
    return graph


def testUnconnected():
    '''
    Creates a graph with two unconnected pipelines.

    :rtype: the created graph
    '''
    graph = WorkflowGraph()
    testPipeline(graph)
    testPipeline(graph)
    return graph


''' important: this is the graph_variable '''
graph = testUnconnected()



# Copyright (c) The University of Edinburgh 2014
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
This is a dispel4py graph that shows the group-by data pattern to count words.
'''

from dispel4py.examples.graph_testing import testing_PEs as t
from dispel4py.workflow_graph import WorkflowGraph


def testGrouping():
    '''
    Creates the test graph.
    '''
    words = t.RandomWordProducer()
    cons1 = t.TestOneInOneOut()
    cons2 = t.TestOneInOneOut()
    cons3 = t.TestOneInOneOut()
    count = t.WordCounter()
    graph = WorkflowGraph()
    graph.connect(words, 'output', cons1, 'input')
    graph.connect(cons1, 'output', cons2, 'input')
    graph.connect(cons2, 'output', cons3, 'input')
    graph.connect(cons3, 'output', count, 'input')

    graph.partitions = [[words], [cons1, cons2, cons3], [count]]
    return graph

graph = testGrouping()



# Copyright (c) The University of Edinburgh 2014
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
This is a dispel4py graph which produces a workflow with a pipeline
in which the producer node ``prod`` sends data to the consumer node ``cons1``
which then sends data to node ``cons2``.
Note that in this graph we have defined several instances of the cons1 and
cons2 nodes and all the instances of the cons1 node are sending data to only
one instance of cons2 node.
This type of grouping is called *global* in dispel4py (all to one).

.. image:: /images/grouping_alltoone.png

It can be executed with MPI and STORM.

* MPI: Please, locate yourself into the dispel4py directory.

    Execute the MPI mapping as follows::

        mpiexec -n <number mpi_processes> dispel4py mpi\\
            [-a name_dispel4py_graph]\\
            [-f file containing the input dataset in JSON format]\\
            [-i number of iterations/runs']\\
            [-s]

    The argument '-s' forces to run the graph in a simple processing, which
    means that the first node of the graph will be executed in a process, and
    the rest of nodes will be executed in a second process.
    When <-i number of interations/runs> is not indicated, the graph is
    executed once by default.

    For example::

        mpiexec -n 11 dispel4py mpi \\
            dispel4py.examples.graph_testing.grouping_alltoone -i 10

    .. note::

        Each node in the graph is executed as a separate MPI process.
        This graph has 3 nodes. For this reason we need at least 3 MPI
        processes to execute it.

    Output::

        Processing 10 iterations.
        Processes: {'TestProducer0': [0], \
'TestOneInOneOut2': [6, 7, 8, 9, 10], 'TestOneInOneOut1': [1, 2, 3, 4, 5]}
        TestProducer0 (rank 0): Processed 10 iterations.
        TestOneInOneOut1 (rank 1): Processed 2 iterations.
        TestOneInOneOut1 (rank 2): Processed 2 iterations.
        TestOneInOneOut1 (rank 4): Processed 2 iterations.
        TestOneInOneOut1 (rank 5): Processed 2 iterations.
        TestOneInOneOut1 (rank 3): Processed 2 iterations.
        TestOneInOneOut2 (rank 7): Processed 0 iterations.
        TestOneInOneOut2 (rank 8): Processed 0 iterations.
        TestOneInOneOut2 (rank 9): Processed 0 iterations.
        TestOneInOneOut2 (rank 6): Processed 10 iterations.
        TestOneInOneOut2 (rank 10): Processed 0 iterations.

    Note that only one instance of the consumer node ``TestOneInOneOut2``
    (rank 6) received all the input blocks from the previous PE,
    as the grouping has been defined as `global`.

* STORM:
'''

from dispel4py.examples.graph_testing import testing_PEs as t
from dispel4py.workflow_graph import WorkflowGraph


def testAlltoOne():
    '''
    Creates a graph with two consumer nodes and a global grouping.

    :rtype: the created graph
    '''
    graph = WorkflowGraph()
    prod = t.TestProducer()
    cons1 = t.TestOneInOneOut()
    cons2 = t.TestOneInOneOut()
    cons1.numprocesses = 5
    cons2.numprocesses = 5
    graph.connect(prod, 'output', cons1, 'input')
    cons2.inputconnections['input']['grouping'] = 'global'
    graph.connect(cons1, 'output', cons2, 'input')
    return graph


''' important: this is the graph_variable '''
graph = testAlltoOne()



# Copyright (c) The University of Edinburgh 2014
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
This is a dispel4py graph which produces a pipeline workflow with one producer
node (prod) and 2 consumer nodes. The second consumer node delays the output by
a fixed time and records the average processing time.

Execution:

* MPI: Please, locate yourself into the dispel4py directory.

    Execute the MPI mapping as follows::

        mpiexec -n <number mpi_processes> python -m dispel4py.worker_mpi\\
               [-a name_dispel4py_graph]\\
               [-f file containing the input dataset in JSON format]\\
               [-i number of iterations/runs']\\
               [-s]

    The argument '-s' forces to run the graph in a simple processing, which
    means that the first node of the graph will be executed in a process, and
    the rest of nodes will be executed in a second process.
    When [-i number of interations/runs] is not indicated, the graph is
    executed once by default.


    For example::

        mpiexec -n 6 dispel4py mpi \\
            dispel4py.examples.graph_testing.delayed_pipeline

    .. note::

        Each node in the graph is executed as a separate MPI process.
        This graph has 3 nodes. For this reason we need at least 3 MPI
        processes to execute it.

    Output::

        Processes: {'TestDelayOneInOneOut2': [2, 3], 'TestProducer0': [4], \
'TestOneInOneOut1': [0, 1]}
        TestProducer0 (rank 4): Processed 10 iterations.
        TestOneInOneOut1 (rank 1): Processed 5 iterations.
        TestOneInOneOut1 (rank 0): Processed 5 iterations.
        TestDelayOneInOneOut2 (rank 3): Average processing time: 1.00058307648
        TestDelayOneInOneOut2 (rank 3): Processed 5 iterations.
        TestDelayOneInOneOut2 (rank 2): Average processing time: 1.00079641342
        TestDelayOneInOneOut2 (rank 2): Processed 5 iterations.

'''

from dispel4py.examples.graph_testing import testing_PEs as t
from dispel4py.workflow_graph import WorkflowGraph
from dispel4py.new.monitoring import ProcessTimingPE

prod = t.TestProducer()
cons1 = t.TestOneInOneOut()
''' adding a processing timer '''
cons2 = ProcessTimingPE(t.TestDelayOneInOneOut())

''' important: this is the graph_variable '''
graph = WorkflowGraph()
graph.connect(prod, 'output', cons1, 'input')
graph.connect(cons1, 'output', cons2, 'input')



# Copyright (c) The University of Edinburgh 2014-2015
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
This is a dispel4py graph which produces a workflow that splits the data and
sends it to two nodes (cons1 and cons2) and the output of those two nodes is
merged by another node (last).

.. image:: /images/split_merge.png

It can be executed with MPI and STORM.

* MPI: Please, locate yourself into the dispel4py directory.

    Execute the MPI mapping as follows::

        mpiexec -n <number mpi_processes> dispel4py mpi <module|module file>\\
            [-a name_dispel4py_graph]\\
            [-f file containing the input dataset in JSON format]\\
            [-i number of iterations/runs']\\
            [-s]

    The argument '-s' forces to run the graph in a simple processing, which
    means that the first node of the graph will be executed in a process, and
    the rest of nodes will be executed in a second process.
    When [-i number of interations/runs] is not indicated, the graph is
    executed once by default.


    For example::

        mpiexec -n 4 dispel4py mpi dispel4py.examples.graph_testing.split_merge

    .. note::

        Each node in the graph is executed as a separate MPI process.
        This graph has 4 nodes. For this reason we need at least 4 MPI
        processes to execute it.

    Output::

        Processing 1 iteration.
        Processes: {'TestProducer0': [1], 'TestOneInOneOutWriter2': [2], \
'TestTwoInOneOut3': [0], 'TestOneInOneOut1': [3]}
        TestProducer0 (rank 1): Processed 1 iteration.
        TestOneInOneOut1 (rank 3): Processed 1 iteration.
        TestOneInOneOutWriter2 (rank 2): Processed 1 iteration.
        TestTwoInOneOut3 (rank 0): Processed 2 iterations.
'''

from dispel4py.examples.graph_testing import testing_PEs as t
from dispel4py.workflow_graph import WorkflowGraph


def testSplitMerge():
    '''
    Creates the split/merge graph with 4 nodes.

    :rtype: the created graph
    '''
    graph = WorkflowGraph()
    prod = t.TestProducer(2)
    cons1 = t.TestOneInOneOut()
    cons2 = t.TestOneInOneOutWriter()
    last = t.TestTwoInOneOut()
    graph.connect(prod, 'output0', cons1, 'input')
    graph.connect(prod, 'output1', cons2, 'input')
    graph.connect(cons1, 'output', last, 'input0')
    graph.connect(cons2, 'output', last, 'input1')
    return graph

''' important: this is the graph_variable '''
graph = testSplitMerge()



# Copyright (c) The University of Edinburgh 2014
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Base PEs implementing typical processing patterns.
'''

from dispel4py.core import GenericPE, NAME


class BasePE(GenericPE):
    '''
    A basic implementation of a :py:class:`~dispel4py.core.GenericPE` that
    allows to easily extend the GenericPE with named inputs and outputs.
    '''
    INPUT_NAME = 'input'
    OUTPUT_NAME = 'output'

    def __init__(self, inputs=[], outputs=[], numInputs=0, numOutputs=0):
        '''
        :param inputs: a list of input names (optional)
        :param outputs: a list of output names (optional)
        :param numInputs: number of inputs; the inputs are generated as
        'input0' to 'input`n`' where `n` is the number of inputs (optional)
        :param numInputs: number of outputs; the outputs are generated as
        'output0' to 'output`n`' where `n` is the number of outputs (optional)
        '''
        GenericPE.__init__(self)

        for i in range(numInputs):
            name = '%s%s' % (BasePE.INPUT_NAME, i)
            self.inputconnections[name] = {NAME: name}
        for i in range(numOutputs):
            name = '%s%s' % (BasePE.OUTPUT_NAME, i)
            self.outputconnections[name] = {NAME: name}
        for name in inputs:
            self.inputconnections[name] = {NAME: name}
        for name in outputs:
            self.outputconnections[name] = {NAME: name}


class IterativePE(BasePE):
    '''
    An iterative PE that has one input and one output stream.
    When executed, this PE produces one output data block for each input block.
    Subclasses are expected to override
    :py:func:`~dispel4py.base.IterativePE._process` to implement processing.
    '''
    INPUT_NAME = 'input'
    OUTPUT_NAME = 'output'

    def __init__(self):
        BasePE.__init__(self)
        self._add_input(IterativePE.INPUT_NAME)
        self._add_output(IterativePE.OUTPUT_NAME)

    def process(self, inputs):
        '''
        Calls the implementation of
        :py:func:`~dispel4py.base.ConsumerPE._process` of the subclass with
        the data read from the input stream and writes the return value to the
        output stream.
        '''
        data = inputs[IterativePE.INPUT_NAME]
        result = self._process(data)
        if result is not None:
            return {self.OUTPUT_NAME: result}

    def _process(self, data):
        '''
        Processes a data block.
        :param data: the input data
        :returns: an output data block or None
        '''
        return None


class ProducerPE(BasePE):
    '''
    A PE that has no input and one output named "output".
    Subclasses are expected to override
    :py:func:`~dispel4py.base.ProducerPE._process` to implement processing.
    '''
    OUTPUT_NAME = 'output'

    def __init__(self):
        BasePE.__init__(self)
        self._add_output(ProducerPE.OUTPUT_NAME)

    def process(self, inputs):
        '''
        Calls the implementation of
        :py:func:`~dispel4py.base.ProducerPE._process` of the subclass.
        '''
        result = self._process()
        if result is not None:
            return {self.OUTPUT_NAME: result}

    def _process(self):
        return None


class ConsumerPE(BasePE):
    '''
    A PE that has one input named "input" and no outputs.
    Subclasses are expected to override
    :py:func:`~dispel4py.base.ConsumerPE._process` to implement processing.
    '''
    INPUT_NAME = 'input'

    def __init__(self):
        BasePE.__init__(self)
        self._add_input(ConsumerPE.INPUT_NAME)

    def process(self, inputs):
        '''
        Calls the implementation of
        :py:func:`~dispel4py.base.ConsumerPE._process` of the subclass with
        the data read from the input stream.
        '''
        data = inputs[ConsumerPE.INPUT_NAME]
        self._process(data)


class SimpleFunctionPE(IterativePE):
    INPUT_NAME = IterativePE.INPUT_NAME
    OUTPUT_NAME = IterativePE.OUTPUT_NAME

    def __init__(self, compute_fn=None, params={}):
        IterativePE.__init__(self)
        if compute_fn:
            self.name = 'PE_%s' % compute_fn.__name__
        self.compute_fn = compute_fn
        self.params = params

    def _process(self, data):
        return self.compute_fn(data, **self.params)


from dispel4py.workflow_graph import WorkflowGraph


def create_iterative_chain(functions,
                           FunctionPE_class=SimpleFunctionPE,
                           name_prefix='PE_',
                           name_suffix=''):

    '''
    Creates a composite PE wrapping a pipeline that processes obspy streams.
    :param chain: list of functions that process data iteratively. The function
    accepts one input parameter, data, and returns an output data block
    (or None).
    :param requestId: id of the request that the stream is associated with
    :param controlParameters: environment parameters for the processing
    elements
    :rtype: dictionary inputs and outputs of the composite PE that was created
    '''

    prev = None
    first = None
    graph = WorkflowGraph()

    for fn_desc in functions:
        try:
            fn = fn_desc[0]
            params = fn_desc[1]
        except TypeError:
            fn = fn_desc
            params = {}

        # print 'adding %s to chain' % fn.__name__
        pe = FunctionPE_class()
        pe.compute_fn = fn
        pe.params = params
        pe.name = name_prefix + fn.__name__ + name_suffix

        if prev:
            graph.connect(prev, IterativePE.OUTPUT_NAME,
                          pe, IterativePE.INPUT_NAME)
        else:
            first = pe
        prev = pe

    # Map inputs and outputs of the wrapper to the nodes in the subgraph
    graph.inputmappings = {'input': (first, IterativePE.INPUT_NAME)}
    graph.outputmappings = {'output': (prev, IterativePE.OUTPUT_NAME)}

    return graph


class CompositePE(WorkflowGraph):
    '''
    Super class for composite PEs. When extending or instantiating this PE
    a function may be provided to populate the graph. For example::

        def create_graph(graph):
            prod = Producer()
            cons = Consumer()
            graph.connect(prod, 'output', cons, 'input')
            graph._map_output('output', cons, 'output')


        class MyCompositePE(CompositePE):
            def __init__(self):
                CompositePE.__init__(self, create_graph)

    This composite PE can be created with the usual command::

        comp = MyCompositeTestPE()

    It is also possible to pass parameters to the composite PE when
    instantiating::

        def create_graph(limit):
            prod = Producer(limit)
            cons = Consumer()
            graph.connect(prod, 'output', cons, 'input')
            graph._map_output('output', cons, 'output')

        class ParameterisedCompositePE(CompositePE):
            def __init__(self, limit)
                CompositePE.__init__(self, create_graph, limit)
    '''
    def __init__(self, create_graph=None, params=None):
        '''
        Instantiate and populate the graph, if the function provided.
        Otherwise, the graph must be populated explicitly by the subclass or
        after instantiating the object.
        '''
        WorkflowGraph.__init__(self)
        self.inputmappings = {}
        self.outputmappings = {}
        if create_graph:
            if params is not None:
                create_graph(self, None)
            else:
                create_graph(self)

    def _map_input(self, input_name, internal_pe, internal_input):
        '''
        Map the composite PE input to the named input of a PE that is contained
        in the graph.
        '''
        self.inputmappings[input_name] = (internal_pe, internal_input)

    def _map_output(self, output_name, internal_pe, internal_output):
        '''
        Map the composite PE output to the named output of a PE that is
        contained in the graph.
        '''
        self.outputmappings[output_name] = (internal_pe, internal_output)



# Copyright (c) The University of Edinburgh 2014
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
This is a dispel4py graph where each MPI process computes a partition of the
workflow instead of a PE instance. This happens automatically when the graph
has more nodes than MPI processes.
In terms of internal execution, the user has control which parts of the graph
are distributed to each MPI process.
See :py:mod:`~test.graph_testing.partition_parallel_pipeline` on how to specify
the partitioning.

.. image:: /images/parallel_pipeline.png

It can be executed with MPI and STORM.

* MPI: Please, locate yourself into the dispel4py directory.

    Execute the MPI mapping as follows::

        mpiexec -n <number mpi_processes> dispel4py mpi\\
            [-a name_dispel4py_graph]\\
            [-f file containing the input dataset in JSON format]\\
            [-i number of iterations/runs']\\
            [-s]

    The argument '-s' forces to run the graph in a simple processing, which
    means that the first node of the graph will be executed in a process, and
    the rest of nodes will be executed in a second process.
    When <-i number of interations/runs> is not indicated, the graph is
    executed once by default.

    For example::

        mpiexec -n 3 dispel4py mpi dispel4py.examples.parallel_pipeline -i 10

    .. note::

        To force the partitioning the graph must have more nodes than available
        MPI processes.
        This graph has 4 nodes and we use 3 MPI processes to execute it.

    Output::

        Processing 10 iterations
        Graph is too large for MPI job size: 4 > 3. Start simple processing.
        Partitions:  [TestProducer0], \
[TestOneInOneOut1, TestOneInOneOut2, TestOneInOneOut3]
        Processes: {'GraphWrapperPE5': [1, 2], 'GraphWrapperPE4': [0]}
        GraphWrapperPE4 (rank 0): I'm a spout
        GraphWrapperPE5 (rank 1): I'm a bolt
        Rank 0: Sending terminate message to [1, 2]
        GraphWrapperPE4 (rank 0): Processed 10 input block(s)
        GraphWrapperPE4 (rank 0): Completed.
        GraphWrapperPE5 (rank 1): Processed 5 input block(s)
        GraphWrapperPE5 (rank 1): Completed.
        GraphWrapperPE5 (rank 2): I'm a bolt
        GraphWrapperPE5 (rank 2): Processed 5 input block(s)
        GraphWrapperPE5 (rank 2): Completed.
'''

from dispel4py.examples.graph_testing import testing_PEs as t
from dispel4py.workflow_graph import WorkflowGraph


def testParallelPipeline():
    '''
    Creates a graph with 4 nodes.

    :rtype: the created graph
    '''
    graph = WorkflowGraph()
    prod = t.TestProducer()
    cons1 = t.TestOneInOneOut()
    cons2 = t.TestOneInOneOut()
    cons3 = t.TestOneInOneOut()

    graph.connect(prod, 'output', cons1, 'input')
    graph.connect(cons1, 'output', cons2, 'input')
    graph.connect(cons1, 'output', cons3, 'input')

    return graph

''' important: this is the graph_variable '''
graph = testParallelPipeline()



# Copyright (c) The University of Edinburgh 2014-2015
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dispel4py.new.mpi_process import process
from dispel4py.workflow_graph import WorkflowGraph
from dispel4py.examples.graph_testing.testing_PEs\
    import TestProducer, TestOneInOneOut

prod = TestProducer()
cons1 = TestOneInOneOut()
cons2 = TestOneInOneOut()

graph = WorkflowGraph()
graph.connect(prod, 'output', cons1, 'input')
graph.connect(cons1, 'output', cons2, 'input')

process(graph, {prod: [{}, {}, {}]})



# Copyright (c) The University of Edinburgh 2014-2015
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Tests for multiprocessing mapping.

Using nose (https://nose.readthedocs.org/en/latest/) run as follows::

    $ nosetests dispel4py/new/multi_process_test.py
    ...
    ----------------------------------------------------------------------
    Ran 3 tests in 0.042s

    OK
'''
import argparse

from dispel4py.examples.graph_testing.testing_PEs\
    import TestProducer, TestOneInOneOut, TestTwoInOneOut
from dispel4py.workflow_graph import WorkflowGraph
from multi_process import process


args = argparse.Namespace
args.num = 5
args.simple = False


def testPipeline():
    prod = TestProducer()
    cons1 = TestOneInOneOut()
    cons2 = TestOneInOneOut()
    graph = WorkflowGraph()
    graph.connect(prod, 'output', cons1, 'input')
    graph.connect(cons1, 'output', cons2, 'input')
    args = argparse.Namespace
    args.num = 5
    args.simple = False
    process(graph, inputs={prod: [{}, {}, {}]}, args=args)


def testSquare():
    graph = WorkflowGraph()
    prod = TestProducer(2)
    cons1 = TestOneInOneOut()
    cons2 = TestOneInOneOut()
    last = TestTwoInOneOut()
    graph.connect(prod, 'output0', cons1, 'input')
    graph.connect(prod, 'output1', cons2, 'input')
    graph.connect(cons1, 'output', last, 'input0')
    graph.connect(cons2, 'output', last, 'input1')
    args.num = 4
    process(graph, inputs={prod: [{}]}, args=args)


def testTee():
    graph = WorkflowGraph()
    prod = TestProducer()
    cons1 = TestOneInOneOut()
    cons2 = TestOneInOneOut()
    graph.connect(prod, 'output', cons1, 'input')
    graph.connect(prod, 'output', cons2, 'input')
    args.num = 3
    process(graph, inputs={prod: [{}, {}, {}, {}, {}]}, args=args)



config = {
    "mpi": "dispel4py.new.mpi_process",
    "multi": "dispel4py.new.multi_process",
    "simple": "dispel4py.new.simple_process",
    "storm": "dispel4py.storm.storm_submission",
    "spark": "dispel4py.new.spark_process"
}



# Copyright (c) The University of Edinburgh 2014-2015
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Collection of dispel4py utilities.
'''

from dispel4py.workflow_graph import WorkflowGraph

from importlib import import_module
from imp import load_source, new_module
import __builtin__
from types import ModuleType
import traceback


class DummyModule(ModuleType):
    def __getattr__(self, key):
        return None
    __all__ = []   # support wildcard imports


def loadGraphIgnoreImports(module_name, graph_var):
    '''
    Loads a graph from the given module and ignores any import errors.
    '''

    def tryimport(name, globals={}, locals={}, fromlist=[], level=-1):
        try:
            return realimport(name, globals, locals, fromlist, level)
        except ImportError as exc:
            print "Warning: %s" % exc
            return DummyModule(name)

    realimport, __builtin__.__import__ = __builtin__.__import__, tryimport

    mod = import_module(module_name)
    graph = getattr(mod, graph_var)

    __builtin__.__import__ = realimport

    return graph


def loadSourceIgnoreImports(module_name, path, attr_name):
    '''
    Import a module from the given source file at 'path' and return the named
    attribute 'attr_name'. Any import errors are being ignored.

    :param module_name: name of the module to load
    :param path: location of the source file
    :param attr_name: name of the attribute within the module
    '''
    def tryimport(name, globals={}, locals={}, fromlist=[], level=-1):
        try:
            return realimport(name, globals, locals, fromlist, level)
        except ImportError as exc:
            print "Warning: %s" % exc
            return DummyModule(name)

    realimport, __builtin__.__import__ = __builtin__.__import__, tryimport

    mod = load_source(module_name, path)
    attr = getattr(mod, attr_name)

    __builtin__.__import__ = realimport

    return attr


def loadSource(module_name, path, attr_name):
    '''
    Import a module from the given source file at 'path' and return
    the named attribute 'attr_name'.

    :param module_name: name of the module to load
    :param path: location of the source file
    :param attr_name: name of the attribute within the module
    '''
    mod = load_source(module_name, path)
    attr = getattr(mod, attr_name)
    return attr


def loadIgnoreImports(module_name, attr_name, code):
    '''
    Import a module from source and return the specified attribute.

    :param module_name: name of the module to load
    :param attr_name: name of the attribute within the module
    :param code: source code of the module
    '''
    def tryimport(name, globals={}, locals={}, fromlist=[], level=-1):
        try:
            return realimport(name, globals, locals, fromlist, level)
        except ImportError as exc:
            print "Warning: %s" % exc
            return DummyModule(name)

    realimport, __builtin__.__import__ = __builtin__.__import__, tryimport

    mod = new_module(module_name)
    exec code in mod.__dict__
    attr = getattr(mod, attr_name)
    __builtin__.__import__ = realimport

    return attr


def findWorkflowGraph(mod, attr):
    if attr is not None:
        # use the named attribute
        graph = getattr(mod, attr)
    else:
        # search for a workflow graph in the given module
        for i in dir(mod):
            attr = getattr(mod, i)
            if isinstance(attr, WorkflowGraph):
                if not hasattr(attr, 'inputmappings')\
                        and not hasattr(attr, 'outputmappings'):
                    graph = attr
    return graph


def loadGraphFromFile(module_name, path, attr=None):
    mod = load_source(module_name, path)
    attr = findWorkflowGraph(mod, attr)
    return attr


def loadGraphFromSource(module_name, source, attr=None):
    mod = new_module(module_name)
    exec source in mod.__dict__
    attr = findWorkflowGraph(mod, attr)
    return attr


def loadGraph(module_name, attr=None):
    '''
    Loads a graph from the given module.
    '''
    mod = import_module(module_name)
    graph = findWorkflowGraph(mod, attr)
    return graph


def load_graph(graph_source, attr=None):
    # try to load from a module
    error_message = ''
    try:
        return loadGraph(graph_source, attr)
    except ImportError:
        # it's not a module
        error_message += 'No module "%s"\n' % graph_source
        pass
    except Exception as e:
        error_message += \
            'Error loading graph module:\n%s' % traceback.format_exc()
        pass

    # maybe it's a file?
    try:
        return loadGraphFromFile('temp', graph_source, attr)
    except IOError:
        # it's not a file
        error_message += 'No file "%s"\n' % graph_source
        pass
    except Exception as e:
        error_message += \
            'Error loading graph from file:\n%s' % traceback.format_exc()

    # or the source code as a string
    try:
        return loadGraphFromSource('temp', graph_source, attr)
    except Exception as e:
        error_message += 'Error trying to parse as Python: %s' % e
        pass

    # we don't know what it is
    print 'Failed to load graph from "%s":\n%s' % (graph_source, error_message)


from sys import getsizeof
from itertools import chain
from collections import deque


def total_size(o, handlers={}, verbose=False):
    """
    From: http://code.activestate.com/recipes/577504/
    Returns the approximate memory footprint an object and all of its contents.

    Automatically finds the contents of the following builtin containers and
    their subclasses:  tuple, list, deque, dict, set and frozenset.
    To search other containers, add handlers to iterate over their contents:

        handlers = {SomeContainerClass: iter,
                    OtherContainerClass: OtherContainerClass.get_elements}

    """
    dict_handler = lambda d: chain.from_iterable(d.items())
    all_handlers = {
        tuple: iter,
        list: iter,
        deque: iter,
        dict: dict_handler,
        set: iter,
        frozenset: iter,
    }
    all_handlers.update(handlers)
    seen = set()
    default_size = getsizeof(0)

    def sizeof(o):
        if id(o) in seen:
            return 0
        seen.add(id(o))
        s = getsizeof(o, default_size)

        for typ, handler in all_handlers.items():
            if isinstance(o, typ):
                s += sum(map(sizeof, handler(o)))
                break
        return s

    return sizeof(o)

import copy


def make_hash(o):

    """
    Makes a hash from a dictionary, list, tuple or set to any level, that
    contains only other hashable types (including any lists, tuples, sets, and
    dictionaries).
    """
    if isinstance(o, (set, tuple, list)):
        return hash(tuple([make_hash(e) for e in o]))

    if not isinstance(o, dict):
        return hash(o)

    new_o = copy.deepcopy(o)
    for k, v in new_o.items():
        new_o[k] = make_hash(v)

    return hash(tuple(frozenset(sorted(new_o.items()))))



import multiprocessing
import pickle
import traceback


class A(object):
    def hello(a):
        def f(x):
            return 2*x
        r = f(a)
        print r
        return r


def _producer():
    a=A()
    try:
        pickle.dumps(a.hello)
        q.put(a.hello)
    except Exception as e:
        print 'Pickling error : %s' % e
        traceback.print_exc()
        pass
    for i in range(5):
        q.put(i)
    q.put(None)

def _consumer():
    d = 0
    while d is not None:
        d = q.get()
        print d


q = multiprocessing.Queue()
p1 = multiprocessing.Process(target=_producer)
p2 = multiprocessing.Process(target=_consumer)


p1.start()
p2.start()

p1.join()
p2.join()


# Copyright (c) The University of Edinburgh 2014-2015
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from mpi4py import MPI

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

from processor import GenericWrapper, simpleLogger
from processor import STATUS_TERMINATED, STATUS_ACTIVE
import processor
import types


def process(workflow, inputs, args):
    processes = {}
    inputmappings = {}
    outputmappings = {}
    success = True
    nodes = [node.getContainedObject() for node in workflow.graph.nodes()]
    if rank == 0 and not args.simple:
        try:
            processes, inputmappings, outputmappings =\
                processor.assign_and_connect(workflow, size)
        except:
            success = False
    success = comm.bcast(success, root=0)

    if args.simple or not success:
        ubergraph = processor.create_partitioned(workflow)
        nodes = [node.getContainedObject() for node in ubergraph.graph.nodes()]
        if rank == 0:
            print 'Partitions: %s' % ', '.join(
                ('[%s]' % ', '.join((pe.id for pe in part))
                    for part in workflow.partitions))
            for node in ubergraph.graph.nodes():
                wrapperPE = node.getContainedObject()
                ns = [n.getContainedObject().id
                      for n in wrapperPE.workflow.graph.nodes()]
                print('%s contains %s' % (wrapperPE.id, ns))

            try:
                processes, inputmappings, outputmappings = \
                    processor.assign_and_connect(ubergraph, size)
                inputs = processor.map_inputs_to_partitions(ubergraph, inputs)
                success = True
            except:
                print 'dispel4py.mpi_process: '\
                      'Not enough processes for execution of graph'
                success = False

    success = comm.bcast(success, root=0)

    if not success:
        return

    inputs = {pe.id: v for pe, v in inputs.iteritems()}
    processes = comm.bcast(processes, root=0)
    inputmappings = comm.bcast(inputmappings, root=0)
    outputmappings = comm.bcast(outputmappings, root=0)
    inputs = comm.bcast(inputs, root=0)

    if rank == 0:
        print 'Processes: %s' % processes
        # print 'Inputs: %s' % inputs

    for pe in nodes:
        if rank in processes[pe.id]:
            provided_inputs = processor.get_inputs(pe, inputs)
            wrapper = MPIWrapper(pe, provided_inputs)
            wrapper.targets = outputmappings[rank]
            wrapper.sources = inputmappings[rank]
            wrapper.process()


import Queue
from threading import Thread


def receive(wrapper):
    while wrapper.terminated < wrapper._num_sources:
        status = MPI.Status()
        msg = comm.recv(source=MPI.ANY_SOURCE, tag=MPI.ANY_TAG, status=status)
        tag = status.Get_tag()
        # print 'Received %s, %s' % (msg, tag)
        if tag == STATUS_TERMINATED:
            wrapper.terminated += 1
        else:
            wrapper.input_data.put((msg, tag))
        # self.wrapper.pe.log('Queue size: %s'%self.wrapper.input_data.qsize())
    # put the final terminate block into the queue
    wrapper.input_data.put((None, STATUS_TERMINATED))


class MPIWrapper(GenericWrapper):

    def __init__(self, pe, provided_inputs=None):
        GenericWrapper.__init__(self, pe)
        self.pe.log = types.MethodType(simpleLogger, pe)
        self.pe.rank = rank
        self.provided_inputs = provided_inputs
        self.terminated = 0
        self.input_data = Queue.Queue()

    def process(self):
        self.reader = Thread(target=receive, args=(self,))
        self.reader.start()
        super(MPIWrapper, self).process()

    def _read(self):
        result = super(MPIWrapper, self)._read()
        if result is not None:
            return result
        return self.input_data.get()

    def _write(self, name, data):
        try:
            targets = self.targets[name]
        except KeyError:
            # no targets
            # self.pe.log('Produced output: %s' % {name: data})
            return
        for (inputName, communication) in targets:
            output = {inputName: data}
            dest = communication.getDestination(output)
            for i in dest:
                # self.pe.log('Sending %s to %s' % (output, i))
                request = comm.isend(output, tag=STATUS_ACTIVE, dest=i)
                status = MPI.Status()
                request.Wait(status)

    def _terminate(self):
        self.reader.join()
        for output, targets in self.targets.iteritems():
            for (inputName, communication) in targets:
                for i in communication.destinations:
                    # self.pe.log('Terminating consumer %s' % i)
                    comm.isend(None, tag=STATUS_TERMINATED, dest=i)



# Copyright (c) The University of Edinburgh 2014
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
The core module for dispel4py.
'''

import uuid
# Connection-level dict elements
NAME = 'name'
TYPE = 'type'
DESCRIPTION = 'descr'
META = 'meta'
GROUPING = 'grouping'
WRITER = 'writer'


class GenericPE(object):
    '''
Base class for Dispel4Py processing elements (PEs). Custom PEs are expected to
extend this class and override the 'process' function.

Custom PEs must override :py:func:`~dispel4py.core.GenericPE.__init__` to
declare the inputs and outputs that can be connected within the workflow
graph, by defining a NAME and possibly a TYPE.
The type of a connection is specific to the enactment system.
In the example below the target system is Storm and the type declares what
kind of tuples are produced::
    out1[TYPE] = ['timestamp', 'origin', 'streams']

In some cases, the output types are determined dynamically and depend on the
input types, for example when implementing a filter which consumes any type
of blocks but the type of the output is the same as the type of the input.
The graph framework supports this by propagating types across the workflow
before enactment and providing each PE with the input types that it can expect
in the method::
    setInputTypes(self, types)
which can be overridden to deduce output types from input types or to raise an
error if the types are not acceptable.
The PE may then override the method::
    getOutputTypes(self)
to declare the output types that it produces.
In the example of a filter PE this method would return the input types.

Custom PEs may implement the method preprocess() to initialise variables or
data before processing commences.

Example implementation::

    import traceback
    import cStringIO
    import base64
    from obspy.core import read,UTCDateTime,Stream,Trace
    from dispel4py.core import GenericPE, NAME, TYPE

    INPUT_NAME = 'input'
    OUTPUT_NAME = 'output'

    class AppendAndSynchronize(GenericPE):

        def __init__(self):
            GenericPE.__init__(self)
            self._add_input(INPUT_NAME)
            self._add_output(OUTPUT_NAME, ['timestamp', 'origin', 'streams'])

        def process(self, inputs):
            values = inputs[INPUT_NAME]
            parameters = values[0]
            origin = values[1]
            data = values[2:]
            if not data:
                self.error+= "No Data";
                raise Exception("No Data!")

            streams=list();
            while data:
                streamItem=data.pop(0);
                streams.append(eval(streamItem["data"]))

            # Reads the first file
            st = read(self.rootpath+"%s" % (streams[0].pop(0),))

            #Reads the following files
            while streams[0]:
                ff= "%s" % (streams[0].pop(0),)
                st += read(self.rootpath+ff)

            starttime="%s" % (parameters["starttime"])
            endtime="%s" % (parameters["endtime"])

            st=st.slice(UTCDateTime(starttime),UTCDateTime(endtime));
            streamtransfer={}
            if type(st) == Stream:
                memory_file = cStringIO.StringIO()
                mseed = st.write(memory_file, format="MSEED")
                streamtransfer={"data":base64.b64encode(memory_file.getvalue())}
            output = [ parameters, origin, streamtransfer ]

            return { OUTPUT_NAME : output }
    '''

    def __init__(self, numprocesses=1):
        self.inputconnections = {}
        self.outputconnections = {}
        self.wrapper = 'simple'
        self.pickleIgnore = []
        self.pickleIgnore = vars(self).keys()
        self.numprocesses = numprocesses

        def log(self, message):
            ''' To be implemented '''
            pass
        self.log = log
        self.name = self.__class__.__name__
        self.id = self.name + str(uuid.uuid4())

    def _add_input(self, name, grouping=None, tuple_type=None):
        '''
        Declares an input for this PE.
        This method may be used when initialising a PE instead of modifying
        :py:attr:`~dispel4py.core.GenericPE.inputconnections` directly.

        :param name: name of the input
        :param grouping: the grouping type that this input expects (optional)
        :param tuple_type: type of tuples accepted by this input (optional)
        '''
        self.inputconnections[name] = {NAME: name}
        if grouping:
            self.inputconnections[name][GROUPING] = grouping
        if tuple_type:
            self.inputconnections[name][TYPE] = tuple_type

    def _add_output(self, name, tuple_type=None):
        '''
        Declares an output for this PE.
        This method may be used when initialising a PE instead of modifying
        :py:attr:`~dispel4py.core.GenericPE.outputconnections` directly.

        :param name: name of the output
        :param tuple_type: type of tuples produced by this output (optional)
        '''
        self.outputconnections[name] = {NAME: name}
        if tuple_type:
            self.outputconnections[name][TYPE] = tuple_type

    def setInputTypes(self, types):
        '''
        Sets the input types of this PE, in the form of a dictionary.
        It is meant to be overridden, e.g. if output types depend on input.

        .. note::

            This method is always called before
            :py:func:`~dispel4py.core.GenericPE.getOutputTypes`.

        :param types: object types for each input stream
        :type types: dictionary mapping input name to input type

        Usage example::

            pe.setInputTypes({'input1':['t1', 't2', 't3'], \
                              'input2':['t4', 't5']})
        '''
        pass

    def getOutputTypes(self):
        '''
        Returns the output types of this PE, in the form of a dictionary.
        This method may be overridden if output types are not static and
        depend on input types.

        .. note::

            This method is only called after the input types have been
            initialised in :py:func:`~dispel4py.core.GenericPE.setInputTypes`.

        :rtype: a dictionary mapping each output name to its type

        By default it returns a dictionary of the types defined in the
        'outputconnections' instance variable.

        Usage example::

            def getOutputTypes(self):
                output = { 'output1' : myInputs['input1'],
                           'output2' : [ 'comment' ] }

        '''
        ret = {}
        # print '%s: %s' % (self.id, self.outputconnections)
        for name, output in self.outputconnections.iteritems():
            try:
                ret[name] = output[TYPE]
            except KeyError:
                raise Exception("%s: No output type defined for '%s'"
                                % (self.id, name))
        return ret

    def _preprocess(self):
        '''
        Subclasses may override this method for variable and data
        initialisation before data processing commences.
        '''
        None

    def preprocess(self):
        '''
        This method called once before processing commences.
        '''
        self._preprocess()

    def _process(self, data):
        '''
        To be overridden by a PE implementation subclass. The data input
        parameter may contain data in any format that is prepared by a PE
        base pattern class implementing
        :py:func:`~dispel4py.core.GenericPE.process`.
        By default, the data parameter contains the dictionary of all inputs
        and the return value must be the output data dictionary.
        PE base classes may override
        :py:func:`~dispel4py.core.GenericPE.process` as they wish to pre and
        postprocess data to and from any format.
        To produce more than one output block data can be written at any point
        during processing using the :py:func:`~dispel4py.core.GenericPE.write`
        method.
        '''
        None

    def process(self, inputs):
        '''
        The 'inputs' dictionary contains data from any or all of the streams
        that are connected to this PE, in any order. The return value of this
        function is a single output dictionary, with the names of the output
        streams as keys.
        PE base classes may override this method to allow preprocessing of data
        before calling :py:func:`~dispel4py.core.GenericPE._process` with the
        prepared data on the PE implementation.
        The data returned by the implementation may then be postprocessed
        before it is returned to the implementation framework.

        :param inputs: the input data for this iteration
        :type inputs: dictionary
        :rtype: a dictionary with the output data
        '''
        return self._process(inputs)

    def _postprocess(self):
        None

    def postprocess(self):
        '''
        This method is called once after the last block has been processed and
        a terminate message was sent to this PE.
        '''
        self._postprocess()

    def write(self, name, data, **kwargs):
        '''
        Allows for preprocessing of data to be written to the output pipe.
        This method should be overridden by PE base classes.
        '''
        self._write(name, data)

    def _write(self, name, data, **kwargs):
        '''
        This writes the 'data' to the output pipe with name 'name' of this PE.
        '''
        try:
            output = self.outputconnections[name]
        except KeyError:
            raise Exception("Can't write data: Unknown output connection\
                            '%s' for PE '%s'" % (name, type(self).__name__))
        output[WRITER].write(data)


class LockstepPE(GenericPE):
    '''
    Representation of a PE which consumes its input in lockstep. The inputs
    dictionary that is passed to the
    :py:func:`~dispel4py.GenericPE.LockStepPE.process` function is guaranteed
    to contain one data item from each of the connected input streams.
    '''

    def __init__(self):
        GenericPE.__init__(self)
        self.wrapper = 'lockstep'


class SourcePE(GenericPE):
    '''
    Representation of data-producing PE, i.e. a PE with no input connections.
    '''

    def __init__(self):
        GenericPE.__init__(self)
        self.wrapper = 'source'



# Copyright (c) The University of Edinburgh 2014
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dispel4py.provenance import ProvenancePE, OUTPUT_DATA, INPUT_NAME
from dispel4py.core import TYPE
import uuid
try:
    from obspy.core.utcdatetime import UTCDateTime
except ImportError:
    pass


class SeismoPE(ProvenancePE):

    INPUT_NAME = INPUT_NAME
    OUTPUT_DATA = OUTPUT_DATA

    def __init__(self):
        ProvenancePE.__init__(self)
        self.outputconnections[OUTPUT_DATA][TYPE] = \
            ['timestamp', 'location', 'streams']

    def getDataStreams(self, inputs):
        values = inputs[INPUT_NAME]
        self._timestamp = values[0]
        self._location = values[1]
        data = values[2:]
        streams = {"streams": data}
        return streams

    def writeOutputStreams(self, outputs):
        output_data = [self._timestamp, self._location]
        for v in outputs["streams"]:
            output_data.append(v)
        result = {OUTPUT_DATA: output_data}
        return result

    def initParameters(self, streams):
        ProvenancePE.initParameters(self, streams)
        self.parameters.update(self._timestamp)
        self.parameters.update(self._location)

    def extractItemMetadata(self, st):
        try:
            streammeta = list()
            for tr in st:
                metadic = {}
                metadic.update({"id": str(uuid.uuid1())})
                for attr, value in tr.stats.__dict__.iteritems():
                    if attr == "mseed":
                        mseed = {}
                        for a, v in value.__dict__.iteritems():
                            try:
                                if type(v) == UTCDateTime:
                                    mseed.update({a: str(v)})
                                else:
                                    mseed.update({a: float(v)})
                            except Exception:
                                mseed.update({a: str(v)})
                        metadic.update({"mseed": mseed})
                    else:
                        try:
                            if type(value) == UTCDateTime:
                                metadic.update({attr: str(value)})
                            else:
                                metadic.update({attr: float(value)})
                        except Exception:
                            metadic.update({attr: str(value)})
                streammeta.append(metadic)
        except Exception:
            streammeta = str(st)
        return streammeta

    def writeStream(self, st, attr):
        streamtransfer = {'data': st, 'attr': attr}
        output_data = [self._timestamp, self._location, streamtransfer]
        self.write(OUTPUT_DATA, output_data)



# Copyright (c) The University of Edinburgh 2014
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from thrift.transport import TSocket
from thrift.transport import TTransport
from thrift.protocol import TBinaryProtocol

from storm import Nimbus


class StormClient:

    '''
    A simple Storm client that connects to the cluster at 'host':'port'.
    '''

    def __init__(self, host, port):
        self.host = host
        self.port = port
        transport = TSocket.TSocket(host, port)
        self.transport = TTransport.TFramedTransport(transport)
        self.protocol = TBinaryProtocol.TBinaryProtocol(self.transport)
        self.client = Nimbus.Client(self.protocol)

    def submitTopology(self, name, uploaded_jar_location, conf, topology):
        '''
        Submits the specified topology with name 'name' to the configured host.

        :param name: name given to the topology
        :param uploaded_jar_location: the path to the location of the jar as \
        created by Storm
        :param conf: configuration of the topology
        :param topology: a Storm topology
        '''
        self.transport.open()
        try:
            self.client.submitTopology(name, uploaded_jar_location,
                                       json.dumps(conf), topology)
        finally:
            self.transport.close()

    def killTopology(self, name):
        '''
        Kills the topology identified by 'name'.
        '''
        self.transport.open()
        try:
            self.client.killTopology(name)
        finally:
            self.transport.close()



'''
Processing elements that implement aggregation functions
(AVG, SUM, COUNT, MIN, MAX).
These are composite PEs that are automatically parallelised if the mapping
supports this.
'''

from dispel4py.workflow_graph import WorkflowGraph
from dispel4py.core import GenericPE
import math


class AggregatePE(GenericPE):
    INPUT_NAME = 'input'
    OUTPUT_NAME = 'output'

    def __init__(self, indexes=[0]):
        GenericPE.__init__(self)
        self._add_input(self.INPUT_NAME)
        self._add_output(self.OUTPUT_NAME)
        self.indexes = indexes
        self.value = [0 for i in indexes]

    def _postprocess(self):
        self.write(AggregatePE.OUTPUT_NAME, self.value)


class ContinuousReducePE(GenericPE):
    INPUT_NAME = 'input'
    OUTPUT_NAME = 'output'

    def __init__(self, indexes=[0]):
        GenericPE.__init__(self)
        self._add_input(self.INPUT_NAME)
        self._add_output(self.OUTPUT_NAME)
        self.indexes = indexes
        self.value = [0 for i in indexes]

    def process(self, inputs):
        self._process(self, inputs)
        self.write(AggregatePE.OUTPUT_NAME, self.value)


class CountPE(AggregatePE):

    def __init__(self):
        AggregatePE.__init__(self, [0])

    def _process(self, inputs):
        self.value = [self.value[0]+1]


class MaxPE(AggregatePE):

    def __init__(self, indexes=[0]):
        AggregatePE.__init__(self, indexes)

    def _process(self, inputs):
        v = inputs[AggregatePE.INPUT_NAME]
        self.value = [max(v[i], self.value[i]) for i in self.indexes]


class MinPE(AggregatePE):

    def __init__(self, indexes=[0]):
        AggregatePE.__init__(self, indexes)
        self.value = [None for i in self.indexes]

    def _process(self, inputs):
        v = inputs[AggregatePE.INPUT_NAME]
        for i in self.indexes:
            self.value[i] = min(v[i], self.value[i])\
                if self.value[i] is not None else v[i]


class SumPE(AggregatePE):

    def __init__(self, indexes=[0]):
        AggregatePE.__init__(self, indexes)

    def _process(self, inputs):
        v = inputs[AggregatePE.INPUT_NAME]
        self.value = [self.value[i]+v[i] for i in self.indexes]


class AverageParallelPE(GenericPE):
    INPUT_NAME = 'input'
    OUTPUT_NAME = 'output'

    def __init__(self, index=0):
        GenericPE.__init__(self)
        self._add_input(self.INPUT_NAME)
        self._add_output(self.OUTPUT_NAME)
        self.index = 0
        self.sum = 0
        self.count = 0

    def _process(self, inputs):
        v = inputs[self.INPUT_NAME][self.index]
        self.sum += v
        self.count += 1

    def _postprocess(self):
        avg = float(self.sum)/self.count
        self.write(self.OUTPUT_NAME, [avg, self.count, self.sum])


class AverageReducePE(GenericPE):
    INPUT_NAME = 'input'
    OUTPUT_NAME = 'output'

    def __init__(self):
        GenericPE.__init__(self)
        self._add_input(self.INPUT_NAME, grouping='global')
        self._add_output(self.OUTPUT_NAME)
        self.index = 0
        self.sum = 0
        self.count = 0

    def _process(self, inputs):
        v = inputs[self.INPUT_NAME]
        self.count += v[1]
        self.sum += v[2]

    def _postprocess(self):
        if self.count != 0:
            avg = float(self.sum)/self.count
            self.write(self.OUTPUT_NAME, [avg, self.count, self.sum])


class StdDevPE(GenericPE):
    INPUT_NAME = 'input'
    OUTPUT_NAME = 'output'

    def __init__(self, index=0):
        GenericPE.__init__(self)
        self._add_input(self.INPUT_NAME)
        self._add_output(self.OUTPUT_NAME)
        self.index = index
        self.sum = 0
        self.sum_squared = 0
        self.count = 0

    def _process(self, inputs):
        v = inputs[self.INPUT_NAME][self.index]
        self.sum += v
        self.sum_squared += v*v
        self.count += 1

    def _postprocess(self):
        std_dev = math.sqrt((self.count * self.sum_squared -
                             self.sum * self.sum) /
                            (self.count * (self.count - 1)))
        self.write(self.OUTPUT_NAME, (std_dev, self.count,
                                      self.sum, self.sum_squared))


class StdDevReducePE(GenericPE):
    INPUT_NAME = 'input'
    OUTPUT_NAME = 'output'

    def __init__(self):
        GenericPE.__init__(self)
        self._add_input(self.INPUT_NAME, grouping='global')
        self._add_output(self.OUTPUT_NAME)
        self.sum = 0
        self.sum_squared = 0
        self.count = 0

    def _process(self, inputs):
        values = inputs[self.INPUT_NAME]
        self.count += values[0]
        self.sum += values[1]
        self.sum_squared += values[2]

    def _postprocess(self):
        std_dev = math.sqrt((self.count * self.sum_squared -
                            self.sum * self.sum) /
                            (self.count * (self.count - 1)))
        self.write(self.OUTPUT_NAME, (std_dev, self.count,
                                      self.sum, self.sum_squared))


def parallel_aggregate(instPE, reducePE):
    composite = WorkflowGraph()
    reducePE.inputconnections[AggregatePE.INPUT_NAME]['grouping'] = 'global'
    reducePE.numprocesses = 1
    composite.connect(instPE, AggregatePE.OUTPUT_NAME,
                      reducePE, AggregatePE.INPUT_NAME)
    composite.inputmappings = {'input': (instPE, AggregatePE.INPUT_NAME)}
    composite.outputmappings = {'output': (reducePE, AggregatePE.OUTPUT_NAME)}
    return composite


def parallelCount():
    '''
    Creates a counter composite PE that is parallelisable using a
    map-reduce pattern.
    The first part of the composite PE is a counter that counts all the inputs,
    the second part sums up the counts of the counter instances.
    The output of this PE is a single value that is the number of input items.
    '''
    pe_sum = SumPE([0])
    pe_sum.name = 'CountReduce'
    return parallel_aggregate(CountPE(), pe_sum)


def parallelSum(indexes=[0]):
    '''
    Creates a SUM composite PE that can be parallelised using a
    map-reduce pattern.
    '''
    return parallel_aggregate(SumPE(indexes), SumPE(indexes))


def parallelMin(indexes=[0]):
    '''
    Creates a MIN composite PE that can be parallelised using a
    map-reduce pattern.
    '''
    return parallel_aggregate(MinPE(indexes), MinPE(indexes))


def parallelMax(indexes=[0]):
    '''
    Creates a MAX composite PE that can be parallelised using a
    map-reduce pattern.
    '''
    return parallel_aggregate(MaxPE(indexes), MaxPE(indexes))


def parallelAvg(index=0):
    '''
    Creates an AVG composite PE that can be parallelised using a
    map-reduce pattern.
    '''
    composite = WorkflowGraph()
    parAvg = AverageParallelPE(index)
    reduceAvg = AverageReducePE()
    composite.connect(parAvg, parAvg.OUTPUT_NAME,
                      reduceAvg, reduceAvg.INPUT_NAME)
    composite.inputmappings = {'input': (parAvg, parAvg.INPUT_NAME)}
    composite.outputmappings = {'output': (reduceAvg, reduceAvg.OUTPUT_NAME)}
    return composite


def parallelStdDev(index=0):
    '''
    Creates a STDDEV composite PE that can be parallelised using a
    map-reduce pattern.
    '''
    composite = WorkflowGraph()
    parStdDev = StdDevPE(index)
    reduceStdDev = StdDevReducePE()
    composite.connect(parStdDev, parStdDev.OUTPUT_NAME,
                      reduceStdDev, reduceStdDev.INPUT_NAME)
    composite.inputmappings = {'input': (parStdDev, parStdDev.INPUT_NAME)}
    composite.outputmappings = {'output':
                                (reduceStdDev, reduceStdDev.OUTPUT_NAME)}
    return composite



# Copyright (c) The University of Edinburgh 2014
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import storm


def stormLogger(message):
    '''
    Logger implementation for Storm.
    '''
    try:
        storm.log(str(message))
    except:
        # just ignore the exception
        pass



# Copyright (c) The University of Edinburgh 2014-2015
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time


class Timer(object):
    def __init__(self, verbose=False):
        self.verbose = verbose

    def __enter__(self):
        self.start = time.time()
        return self

    def __exit__(self, *args):
        self.end = time.time()
        self.secs = self.end - self.start


class MonitoringWrapper(object):

    def __init__(self, baseObject):
        self.__class__ = type(baseObject.__class__.__name__,
                              (self.__class__, baseObject.__class__),
                              {})
        self.__dict__ = baseObject.__dict__
        self.baseObject = baseObject


class ReadTimingWrapper(MonitoringWrapper):

    def __init__(self, baseObject):
        MonitoringWrapper.__init__(self, baseObject)
        self.readtime = None
        self.readrate = []

    def _read(self):
        now = time.time()
        if self.readtime:
            self.readrate.append(now-self.readtime)
        self.readtime = now
        return self.baseObject._read()

    def _terminate(self):
        self.log("Average read rate : %s" % (sum(self.readrate) /
                                             float(len(self.readrate))))
        self.baseObject._terminate()


class ProcessTimingPE(MonitoringWrapper):

    def __init__(self, baseObject):
        MonitoringWrapper.__init__(self, baseObject)
        self.times_total = 0
        self.times_count = 0

    def process(self, inputs):
        with Timer() as t:
            result = self.baseObject.process(inputs)
        self.times_total += t.secs
        self.times_count += 1
        return result

    def _postprocess(self):
        self.log('Average processing time: %s' % (self.times_total /
                                                  self.times_count))



# Copyright (c) The University of Edinburgh 2014
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Submits a dispel4py graph for processing to Storm.
All dependencies must be available in the named resources directory or from the
registry.

From the commandline, run the following::

    dispel4py storm <module> [name] [-h] -m {local,remote,create}
                             [-r resourceDir] [-a attribute] [-s]

with positional arguments:

:module: module that creates a dispel4py graph
:name:   name of Storm topology to submit (optional)

and optional arguments:

-h, --help
    show this help message and exit
-m, --mode mode
    execution mode, one of {local, remote, create}
-r, --resources resourceDir
    path to local modules used by the graph - default "./resources/" (optional)
-a, --attr attribute
    name of graph variable in the module (optional)
-s, --save
    do not remove Storm resources after submission (default is to remove
    resources)

.. note::
    A Storm topology, once submitted, runs forever until it is explicitly
    killed.
'''

import argparse
import datetime
import getpass
import os
import shutil
import subprocess
import sys
import tempfile
import traceback

from thrift.transport import TTransport
from thrift.protocol import TBinaryProtocol

from dispel4py.storm.topology import buildTopology

TOPOLOGY_THRIFT_FILE = 'topology.thrift'
STORM_SUBMISSION_CLIENT = 'storm_submission_client.py'

ROOT_DIR = os.path.abspath(os.path.join(os.path.join(
    os.path.join(__file__, os.pardir), os.pardir), os.pardir)) + '/'


def _mkdir_ifnotexists(path):
    try:
        os.mkdir(path)
    except OSError:
        pass


def createPackage(graph, args, static_input):
    '''
    Creates a Storm submission package for the given dispel4py graph.

    :param module_name: name of the graph module that creates a graph
    :param attr: name of the graph attribute within the module - if None the
    first WorkflowGraph is used
    :param res: resource directory - if None the default is "resources"
    :rtype: name of the temporary directory that contains the submission
    package
    '''
    print ROOT_DIR

    res = args.resources
    if res is None:
        res = 'resources'

    # create a temporary directory
    tmpdir = tempfile.mkdtemp()
    tmp_resources_dir = tmpdir + '/' + res

    # copy dependencies of PEs in the graph to resources in temp directory
    shutil.copytree(res, tmp_resources_dir)
    tmp_dispel4py_dir = tmp_resources_dir + '/dispel4py'
    _mkdir_ifnotexists(tmp_dispel4py_dir)
    _mkdir_ifnotexists(tmp_dispel4py_dir + '/storm')
    shutil.copy(ROOT_DIR + 'dispel4py/__init__.py', tmp_dispel4py_dir)
    shutil.copy(ROOT_DIR + 'dispel4py/core.py', tmp_dispel4py_dir)
    shutil.copy(ROOT_DIR + 'dispel4py/base.py', tmp_dispel4py_dir)
    shutil.copy(ROOT_DIR + 'dispel4py/workflow_graph.py', tmp_dispel4py_dir)
    shutil.copy(ROOT_DIR + 'dispel4py/storm/__init__.py',
                tmp_dispel4py_dir + '/storm/')
    shutil.copy(ROOT_DIR + 'dispel4py/storm/utils.py',
                tmp_dispel4py_dir + '/storm/')

    # copy client and dependencies for storm submission to the temp directory
    dispel4py_dir = tmpdir + '/dispel4py'
    _mkdir_ifnotexists(dispel4py_dir)
    _mkdir_ifnotexists(dispel4py_dir + '/storm')
    shutil.copy(ROOT_DIR + 'dispel4py/__init__.py', dispel4py_dir)
    shutil.copy(ROOT_DIR + 'dispel4py/__init__.py', dispel4py_dir + '/storm/')
    shutil.copy(ROOT_DIR + 'dispel4py/storm/client.py',
                dispel4py_dir + '/storm/')
    shutil.copy(ROOT_DIR + 'dispel4py/storm/storm_submission_client.py',
                tmpdir)
    shutil.copy(ROOT_DIR + 'java/src/dispel4py/storm/ThriftSubmit.java',
                tmpdir + '/dispel4py/storm/')
    try:
        shutil.copytree(ROOT_DIR + 'storm', tmpdir + '/storm')
    except:
        pass

    sources = []
    for node in graph.graph.nodes():
        pe = node.getContainedObject()
        is_source = True
        for edge in graph.graph.edges(node, data=True):
            if pe == edge[2]['DIRECTION'][1]:
                is_source = False
                break
        if is_source:
            sources.append(pe)
    print "Sources: %s" % [s.id for s in sources]
    for pe in sources:
        pe._static_input = static_input[pe.id]
        pe._num_iterations = args.iter

    # create the storm topology
    topology = buildTopology(graph)

    # cache PE dependencies imported from the registry to resources_dir
    # registry.createResources(resources_dir, registry.currentRegistry())

    # write thrift representation of the topology to a file
    transportOut = TTransport.TMemoryBuffer()
    protocolOut = TBinaryProtocol.TBinaryProtocol(transportOut)
    topology.write(protocolOut)
    bytes = transportOut.getvalue()
    with open(tmpdir+'/'+TOPOLOGY_THRIFT_FILE, "w") as thrift_file:
        thrift_file.write(bytes)
    return tmpdir


def _getStormHome():
    try:
        STORM_HOME = os.environ['STORM_HOME']
        return STORM_HOME
    except KeyError:
        print 'Error: Please provide the installation directory of Storm as \
               environment variable STORM_HOME'
        sys.exit(1)


def submit(workflow, args, inputs):
    '''
    Creates a Storm submission package and submits it to a remote cluster.

    :param mod: module that creates a dispel4py graph
    :param attr: name of graph attribute within the module - if None the first
    WorkflowGraph is used
    :param res: resource directory
    :param save: if True the Storm submission package is not deleted at the end
    of the run
    '''
    STORM_HOME = _getStormHome()
    topologyName = args.name
    tmpdir = createPackage(workflow, args, inputs)
    print 'Created Storm submission package in %s' % tmpdir
    try:
        print "Submitting topology '%s' to Storm" % topologyName
        stormshell = '%s/bin/storm' % STORM_HOME, 'shell',\
                     'resources/',\
                     'python', 'storm_submission_client.py',\
                     topologyName
        proc = subprocess.Popen(stormshell, cwd=tmpdir)
        proc.wait()
    except:
        pass
    if args.save:
        print tmpdir
    else:
        shutil.rmtree(tmpdir)
        print 'Deleted %s' % tmpdir


def runLocal(workflow, args, inputs):
    '''
    Creates a Storm submission package and executes it locally.
    Note that the Storm topology runs until the process is explicitly killed,
    for example by pressing Ctrl-C.

    :param mod: module that creates a dispel4py graph
    :param attr: name of graph attribute within the module - if None the first
    WorkflowGraph is used
    :param res: resource directory
    :param save: if True the Storm submission package is not deleted at the
    end of the run
    '''
    STORM_HOME = _getStormHome()
    topologyName = args.name
    tmpdir = createPackage(workflow, args, inputs)
    print 'Created Storm submission package in %s' % tmpdir
    try:
        print 'Compiling java client'
        javacp = '.:%s/lib/*:%s/*' % (STORM_HOME, STORM_HOME)
        javacmd = 'javac', '-cp', javacp, 'dispel4py/storm/ThriftSubmit.java'
        proc = subprocess.Popen(javacmd, cwd=tmpdir)
        proc.wait()
        print 'Running topology in local mode'
        javacmd = 'java', '-cp', javacp,\
                  'dispel4py.storm.ThriftSubmit',\
                  'topology.thrift',\
                  topologyName
        proc = subprocess.Popen(javacmd, cwd=tmpdir)
        proc.wait()
    except:
        print traceback.format_exc()
        if args.save:
            print tmpdir
        else:
            shutil.rmtree(tmpdir)
            print 'Deleted %s' % tmpdir


def create(workflow, args, inputs):
    '''
    Creates a Storm submission package and prints the temp directory containing
    the package.

    :param mod: module that creates a dispel4py graph
    :param attr: name of graph attribute within the module - if None
    the first WorkflowGraph is used
    :param res: resource directory
    '''
    tmpdir = createPackage(workflow, args, inputs)
    print 'Created Storm submission package in %s' % tmpdir


def process(workflow, inputs, args=None):
    if args.mode == 'local':
        runLocal(workflow, args, inputs)
    elif args.mode == 'remote':
        submit(workflow, args, inputs)
    elif args.mode == 'create':
        create(workflow, args, inputs)


def parse_args(args, namespace):
    parser = argparse.ArgumentParser(
        description='Submit a dispel4py graph for processing to Storm.')
    defaultName = getpass.getuser() + '_' + \
        datetime.datetime.today().strftime('%Y%m%dT%H%M%S')
    parser.add_argument('name', nargs='?', default=defaultName,
                        help='name of Storm topology to submit')
    parser.add_argument('-m', '--mode',
                        help='execution mode',
                        choices=['local', 'remote', 'create'],
                        required=True)
    parser.add_argument('-r', '--resources', metavar='resourceDir',
                        help='path to local modules used by the graph \
                              - default "./resources/" ')
    parser.add_argument('-s', '--save',
                        help='do not remove Storm resources after submission',
                        action='store_true')
    result = parser.parse_args(args, namespace)
    return result


# if __name__ == "__main__":
#     parser = get_arg_parser()
#     args = parser.parse_args()
#
#     inputs = None
#     num_iterations = None
#     if args.file:
#         try:
#             with open(args.file) as inputfile:
#                 inputs = json.loads(inputfile.read())
#             print("Processing input file %s" % args.file)
#             if type(inputs) != list:
#                 inputs = [inputs]
#         except:
#             print traceback.format_exc()
#             print('Cannot read input file %s' % args.file)
#             sys.exit(1)
#     elif args.iter > 0:
#         #inputs = [ {} for i in range(args.iter) ]
#         print("Processing %s iterations" % args.iter)
#         num_iterations = args.iter
#
#     if args.mode == 'local':
#         runLocal(args, inputs, num_iterations)
#     elif args.mode == 'remote':
#         submit(args, inputs, num_iterations)
#     elif args.mode == 'create':
#         create(args)
#



# Copyright (c) The University of Edinburgh 2014-2015
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
This is a dispel4py graph which produces a pipeline workflow with one producer
node (prod) and 5 consumer nodes.
It can be executed with MPI and STORM.

.. image:: /images/pipeline_test.png

Execution:

* MPI: Please, locate yourself into the dispel4py directory.

    Execute the MPI mapping as follows::

        mpiexec -n <number mpi_processes> dispel4py mpi\\
            [-a name_dispel4py_graph]\\
            [-f file containing the input dataset in JSON format]\\
            [-i number of iterations/runs']\\
            [-s]

    The argument '-s' forces to run the graph in a simple processing, which
    means that the first node of the graph will be executed in a process, and
    the rest of nodes will be executed in a second process.
    When [-i number of interations/runs] is not indicated, the graph is
    executed once by default.


    For example::

        mpiexec -n 6 dispel4py mpi\\
            dispel4py.examples.graph_testing.pipeline_test

    .. note::

        Each node in the graph is executed as a separate MPI process.
        This graph has 6 nodes.
        For this reason we need at least 6 MPI processes to execute it.

    Output::

        Processing 10 iterations.
        Processes: {'TestProducer0': [5], 'TestOneInOneOut5': [2], \
'TestOneInOneOut4': [4], 'TestOneInOneOut3': [3], 'TestOneInOneOut2': [1], \
'TestOneInOneOut1': [0]}
        TestProducer0 (rank 5): Processed 10 iterations.
        TestOneInOneOut1 (rank 0): Processed 10 iterations.
        TestOneInOneOut2 (rank 1): Processed 10 iterations.
        TestOneInOneOut3 (rank 3): Processed 10 iterations.
        TestOneInOneOut4 (rank 4): Processed 10 iterations.
        TestOneInOneOut5 (rank 2): Processed 10 iterations.

* STORM:

    From the dispel4py directory launch the Storm submission client::

        dispel4py storm dispel4py.examples.graph_testing.pipeline_test\\
            -m remote

    Output::

        Spec'ing TestOneInOneOut1
        Spec'ing TestOneInOneOut2
        Spec'ing TestOneInOneOut3
        Spec'ing TestOneInOneOut4
        Spec'ing TestOneInOneOut5
        Spec'ing TestProducer6
        spouts {'TestProducer6': ... }
        bolts  {'TestOneInOneOut5': ... }
        Created Storm submission package in \
/var/folders/58/7bjr3s011kgdtm5lx58prc_40000gn/T/tmp5ePEq3
        Running: java -client -Dstorm.options= -Dstorm.home= ...
        Submitting topology 'TestTopology' to storm.example.com:6627 ...

'''

from dispel4py.examples.graph_testing import testing_PEs as t
from dispel4py.workflow_graph import WorkflowGraph


def testPipeline(graph):
    '''
    Adds a pipeline to the given graph.

    :rtype: the created graph
    '''
    prod = t.TestProducer()
    prev = prod
    for i in range(5):
        cons = t.TestOneInOneOut()
        graph.connect(prev, 'output', cons, 'input')
        prev = cons
    return graph
''' important: this is the graph_variable '''
graph = testPipeline(WorkflowGraph())



from dispel4py.base import IterativePE

INPUT_NAME = IterativePE.INPUT_NAME
OUTPUT_NAME = IterativePE.OUTPUT_NAME


class SeismoStreamPE(IterativePE):

    INPUT_NAME = IterativePE.INPUT_NAME
    OUTPUT_NAME = IterativePE.OUTPUT_NAME

    def __init__(self, compute_fn, params={}):
        IterativePE.__init__(self)
        self.compute_fn = compute_fn
        self.params = params

    def _reset(self):
        self._inputs = None
        self._input_tuple = None
        self._data = None
        self._metadata = None
        self._timestamp = None
        self._location = None

    def _assign(self, inputs):
        self._inputs = inputs[SeismoStreamPE.INPUT_NAME]
        self._input_tuple = self._inputs[2:]
        self._data = [d.data for d in self._input_data]
        self._metadata = [d.metadata for d in self._input_data]
        self._timestamp = self._inputs[0]
        self._location = self._inputs[1]
        self._output_tuple = []

    def get_data(self):
        return self._input_data

    def write_data(self, data, metadata=None):
        output = [self._timestamp, self._location, DataTuple(data, metadata)]
        self.write(SeismoStreamPE.OUTPUT_NAME, output)

    def process(self, inputs):
        try:
            self._assign(inputs)
            result = self.compute_fn(self, **self.params)
            if result:
                self.write_data(result)
        finally:
            self._reset()


class DataTuple(object):
    def __init__(self, data=None, metadata=None):
        self.data = data
        self.metadata = metadata

from dispel4py.workflow_graph import WorkflowGraph


def create_pipeline(chain, name_prefix='SeismoStreamPE_', name_suffix=''):
    '''
    Creates a composite PE wrapping a pipeline that processes obspy streams.
    :param chain: list of functions that process obspy streams.\
        The function takes one input parameter, stream,\
        and returns an output stream.
    :param requestId: id of the request that the stream is associated with
    :param controlParameters: environment parameters for the processing\
        elements
    :rtype: dictionary inputs and outputs of the composite PE that was created
    '''
    prev = None
    first = None
    graph = WorkflowGraph()

    for fn_desc in chain:
        try:
            fn = fn_desc[0]
            params = fn_desc[1]
        except TypeError:
            fn = fn_desc
            params = {}

        pe = SeismoStreamPE(fn, params)
        pe.name = name_prefix + fn.__name__ + name_suffix

        if prev:
            graph.connect(prev, SeismoStreamPE.OUTPUT_NAME,
                          pe, SeismoStreamPE.INPUT_NAME)
        else:
            first = pe
        prev = pe

    # Map inputs and outputs of the wrapper to the nodes in the subgraph
    graph.inputmappings = {'input': (first, INPUT_NAME)}
    graph.outputmappings = {'output': (prev, OUTPUT_NAME)}

    return graph



# Copyright (c) The University of Edinburgh 2014-2015
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
This is a dispel4py graph which produces a workflow that sends copies of the
output data from the producer node (words) to two nodes (filter1 and filter2),
and the outputs of those two filters are merged in the last node (count).

.. image:: /images/grouping_split_merge.png

It can be executed with MPI and STORM.

* MPI: Please, locate yourself into the dispel4py directory.

    Execute the MPI mapping as follows::

        mpiexec -n <number mpi_processes> python -m dispel4py.worker_mpi\\
            [-a name_dispel4py_graph]\\
            [-f file containing the input dataset in JSON format]\\
            [-i number of iterations/runs']\\
            [-s]

    The argument '-s' forces to run the graph in a simple processing, which
    means that the first node of the graph will be executed in a process, and
    the rest of nodes will be executed in a second process.
    When <-i number of interations/runs> is not indicated, the graph is
    executed once by default.

    For example::

        mpiexec -n 4 python -m dispel4py.worker_mpi\\
            dispel4py.examples.graph_testing.grouping_split_merge


    .. note::

        Each node in the graph is executed as a separate MPI process.
        This graph has 4 nodes. For this reason we need at least 4 MPI
        processes to execute it.

    Output::

        Processing 1 iterations
        Processes: {'RandomFilter2': [3], 'WordCounter3': [1], \
'RandomFilter1': [0], 'RandomWordProducer0': [2]}
        RandomFilter1 (rank 0): I'm a bolt
        RandomWordProducer0 (rank 2): I'm a spout
        WordCounter3 (rank 1): I'm a bolt
        Rank 2: Sending terminate message to [0]
        Rank 2: Sending terminate message to [3]
        RandomWordProducer0 (rank 2): Processed 1 input block(s)
        RandomWordProducer0 (rank 2): Completed.
        RandomFilter2 (rank 3): I'm a bolt
        Rank 3: Sending terminate message to [1]
        RandomFilter2 (rank 3): Processed 1 input block(s)
        RandomFilter2 (rank 3): Completed.
        Rank 0: Sending terminate message to [1]
        RandomFilter1 (rank 0): Processed 1 input block(s)
        RandomFilter1 (rank 0): Completed.
        WordCounter3 (rank 1): Processed 2 input block(s)
        WordCounter3 (rank 1): Completed.

    .. note::

        As those PEs are filtering randomly the output could be completely
        different.
'''

from dispel4py.examples.graph_testing import testing_PEs as t
from dispel4py.workflow_graph import WorkflowGraph


def testGrouping():
    '''
    Creates the test graph.
    '''
    words = t.RandomWordProducer()
    filter1 = t.RandomFilter()
    filter2 = t.RandomFilter()
    count = t.WordCounter()
    graph = WorkflowGraph()
    graph.connect(words, 'output', filter1, 'input')
    graph.connect(words, 'output', filter2, 'input')
    graph.connect(filter1, 'output', count, 'input')
    graph.connect(filter2, 'output', count, 'input')

    return graph

''' important: this is the graph_variable '''
graph = testGrouping()



# Copyright (c) The University of Edinburgh 2014
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dispel4py.seismo.seismo import SeismoPE
import traceback

INPUT_NAME = 'input'
OUTPUT_NAME = 'output'


class ObspyStreamPE(SeismoPE):
    '''
    A SeismoPE that calls a function to process an input stream.
    '''

    def __init__(self):
        SeismoPE.__init__(self)

    def setCompute(self, compute_fn, params={}):
        '''
        Define the compute function that this PE uses for processing input
        streams, and any input parameters for the function.
        The function must have at least one input, an obspy stream, and can
        accept more input parameters that must be provided before the PE is
        executed.
        '''
        self.compute_fn = compute_fn, dict(params)

    def setInputTypes(self, types):
        self.inout_types = {OUTPUT_NAME: types[INPUT_NAME]}

    def getOutputTypes(self):
        # output = input
        return self.inout_types

    def compute(self):
        '''
        Calls the processing function with the given parameters and one input
        stream.
        '''
        try:
            try:
                func, params = self.compute_fn
            except TypeError:
                func = self.compute_fn
                params = {}
            output = func(self, self.st, **params)
            self.outputstreams.append(output)
        except:
            self.log(traceback.format_exc())
            self.error += traceback.format_exc()
            self.log("Failed to execute function '%s' with parameters %s"
                     % (func.__name__, params))

from dispel4py.workflow_graph import WorkflowGraph


def createProcessingComposite(chain, suffix='',
                              controlParameters={},
                              provRecorder=None):
    '''
    Creates a composite PE wrapping a pipeline that processes obspy streams.
    :param chain: list of functions that process obspy streams. The function
    takes one input parameter, stream, and returns an output stream.
    :param requestId: id of the request that the stream is associated with
    :param controlParameters: environment parameters for the processing\
        elements
    :rtype: dictionary inputs and outputs of the composite PE that was\
        created
    '''
    prev = None
    first = None
    graph = WorkflowGraph()

    for fn_desc in chain:
        pe = ObspyStreamPE()
        try:
            fn = fn_desc[0]
            params = fn_desc[1]
        except TypeError:
            fn = fn_desc
            params = {}

        pe.compute_fn = fn
        pe.name = 'ObspyStreamPE_' + fn.__name__ + suffix
        pe.controlParameters = controlParameters
        pe.appParameters = dict(params)
        pe.setCompute(fn, params)

        # connect the metadata output to the provenance recorder PE
        # if there is one
        if provRecorder:
            graph.connect(pe, 'metadata', provRecorder, 'metadata')

        if prev:
            graph.connect(prev, OUTPUT_NAME, pe, INPUT_NAME)
        else:
            first = pe
        prev = pe

    # Map inputs and outputs of the wrapper to the nodes in the subgraph
    graph.inputmappings = {'input': (first, INPUT_NAME)}
    graph.outputmappings = {'output': (prev, OUTPUT_NAME)}

    return graph



from dispel4py.examples.graph_testing.testing_PEs \
    import TestProducer, TestOneInOneOut

from dispel4py.base import CompositePE, IterativePE


class MultiplyPE(IterativePE):

    def __init__(self, parameter):
        IterativePE.__init__(self)
        self.parameter = parameter

    def _process(self, data):
        return self.parameter * data


def create_test_graph(graph):
    prod = TestProducer()
    cons = TestOneInOneOut()
    graph.connect(prod, 'output', cons, 'input')
    graph._map_output('output', cons, 'output')


class MyCompositeTestPE(CompositePE):
    def __init__(self):
        CompositePE.__init__(self)
        create_test_graph(self)


def create_graph_param(graph, param):
    cons1 = TestOneInOneOut()
    cons2 = MultiplyPE(param)
    graph.connect(cons1, 'output', cons2, 'input')
    graph._map_input('input', cons1, 'input')
    graph._map_output('output', cons2, 'output')


class MyParameterisedCompositePE(CompositePE):
    def __init__(self, param):
        CompositePE.__init__(self)
        create_graph_param(self, param)


from dispel4py.workflow_graph import WorkflowGraph

graph = WorkflowGraph()
composite = MyCompositeTestPE()
comp2 = MyParameterisedCompositePE(5)
cons1 = TestOneInOneOut()
graph.connect(composite, 'output', comp2, 'input')
graph.connect(comp2, 'output', cons1, 'input')



# Copyright (c) The University of Edinburgh 2014-2015
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import argparse
from importlib import import_module


def main(args=None):

    parser = argparse.ArgumentParser(
        description='Submit a dispel4py graph for processing.')
    parser.add_argument('target', help='target execution platform')
    args, remaining = parser.parse_known_args()
    try:
        from dispel4py.new import mappings
        # see if platform is in the mappings file as a simple name
        target = mappings.config[args.target]
    except KeyError:
        # it is a proper module name - fingers crossed...
        target = args.target
    try:
        process = getattr(import_module(target), 'main')
    except:
        # print traceback.format_exc()
        print 'Unknown target: %s' % target
        return
    process()


if __name__ == "__main__":
    main()
